// Copyright (C) 2024  Umorpha Systems
// SPDX-License-Identifier: AGPL-3.0-or-later

package main

import (
	"context"
	"fmt"
	"os"
	"time"

	"github.com/datawire/ocibuild/pkg/cliutil"
	"github.com/spf13/cobra"

	"git.mothstuff.lol/lukeshu/eclipse/lib/gitcache"
)

func main() {
	argparser := &cobra.Command{
		Use:   os.Args[0] + " {[flags]|SUBCOMMAND...}",
		Short: "Directly interact with a gitcache",

		Args: cliutil.WrapPositionalArgs(cliutil.OnlySubcommands),
		RunE: cliutil.RunSubcommands,

		SilenceErrors: true, // we'll handle this ourselves after .ExecuteContext()
		SilenceUsage:  true, // FlagErrorFunc will handle this

		CompletionOptions: cobra.CompletionOptions{
			DisableDefaultCmd: true,
		},
	}
	argparser.SetFlagErrorFunc(cliutil.FlagErrorFunc)
	argparser.SetHelpTemplate(cliutil.HelpTemplate)

	var cache gitcache.Cache
	argparser.PersistentFlags().StringVarP(&cache.Dir, "cache-dir", "d", "/var/cache/gitcache.git",
		"Where to store cached data")
	argparser.PersistentFlags().DurationVar(&cache.MinPeriod, "min-interval", 5*time.Second,
		"Do not fetch from the same remote more often than this")

	argparser.AddCommand(&cobra.Command{
		Use:   "fetch URL",
		Short: "Update the cache's store of the repository at URL",
		Args:  cliutil.WrapPositionalArgs(cobra.ExactArgs(1)),
		RunE: func(_ *cobra.Command, args []string) error {
			return cache.Fetch(os.Stderr, args[0])
		},
	})
	argparser.AddCommand(&cobra.Command{
		Use:   "rev-parse URL REV",
		Short: "Resolve REF to an object ID within the repository at URL",
		Args:  cliutil.WrapPositionalArgs(cobra.ExactArgs(2)),
		RunE: func(_ *cobra.Command, args []string) error {
			objID, err := cache.RevParse(os.Stderr, args[0], args[1])
			if err != nil {
				return err
			}
			fmt.Println(objID)
			return nil
		},
	})
	argparser.AddCommand(&cobra.Command{
		Use:   "clone URL DIR [FLAGS...]",
		Short: "Clone the repository at URL from the cache to DIR",
		Args:  cliutil.WrapPositionalArgs(cobra.MinimumNArgs(2)),
		RunE: func(cmd *cobra.Command, args []string) error {
			return cache.Clone(os.Stderr, args[0], args[1], args[2:]...)
		},
	})

	ctx := context.Background()

	if err := argparser.ExecuteContext(ctx); err != nil {
		fmt.Fprintf(os.Stderr, "%v: error: %v\n", argparser.CommandPath(), err)
		os.Exit(1)
	}
}
